/* SMG2 serial reader; (C)2012-2014 Jaroslav Fojtik, Licence GPL. */


#include <stdio.h>          /* Standard input/output definitions */
#include <stdlib.h>
#include <string.h>

#include <signal.h>
#include <time.h>


#ifdef _WIN32
  #include <windows.h>
  #define DEFAULT_DEVICE		"COM1"
  #define DEFAULT_STAT_FILE		"C:\\temp\\SMG2.txt"
  #define CRTSCTS			0  
#else
  #include <termios.h>  
  #include <unistd.h>
  #define DEFAULT_DEVICE		"/dev/ttyUSB0"
  #define DEFAULT_STAT_FILE		"/tmp/SMG2.txt"  
#endif


#include "serial.h"


unsigned MAX_RRD_RATE_MS = 10000;		/**< Maximal RRD update rate 1s */
long SerialTimeout_us = 500000;			/**< 0.5s serial line timeout. */
#define MINIMAL_TIMEOUT_us 100000		/**< 0.1s minimal timeout. */
int RequestCount;

int CRtsCts = 0;


/* baudrate settings are defined in <asm/termbits.h>, which is
   included by <termios.h> */
#define BAUDRATE 115200
/* change this definition for the correct port */
const char *MODEMDEVICE = DEFAULT_DEVICE;


const char *rrdupdate_PATH = "/usr/sbin/rrdupdate";
const char *RRD_DATABASE = "/var/rrd/radiation.rrd";
const char *LOG_FILE = NULL;

const char *SMG2_STAT_FILE = DEFAULT_STAT_FILE;
FILE *LogFile = NULL;

int Verbosity = 0;
#define FILTER_FACTOR	0.98

T_COM F = INVALID_HANDLE;		/** Serial SMG2 line. */


static volatile int SigTerm = 0;	///< Counter of sigterm signals. 2'nd sigterm calls abort.
static volatile int SigUsrCount = 0;	///< Counter of user signal.

int sock = -1;


#define CHK(x) x==NULL?"NULL":x



const char *FormatTime(void)
{
static char msg[64];

#if defined(_WIN32) && !defined(__MINGW32__)
    SYSTEMTIME  t;
      
    GetLocalTime(&t);		/* add time to message */  
    sprintf(msg, ("%04d/%02d/%02d %02d:%02d:%02d.%03d "),
      t.wYear, t.wMonth, t.wDay, t.wHour, t.wMinute, t.wSecond, t.wMilliseconds);
#else
 #ifdef __MINGW32__
    timeval t1;
    gettimeofday(&t1, NULL);
 #else
    struct timespec current_ntime;
    clock_gettime(CLOCK_REALTIME,&current_ntime);
 #endif
    time_t current_time;
    const struct tm *t;

    current_time=time((time_t *) NULL);
    t=localtime(&current_time);

    sprintf(msg, ("%04d/%02d/%02d %02d:%02d:%02d.%03d "),
      //mesice jsou cislovany od nuly, musime zobrazovat cislo o 1 vetsi
      t->tm_year+1900, t->tm_mon+1, t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec,
 #ifdef __MINGW32__
      (int)(t1.tv_usec/1000));
 #else
      (int)(current_ntime.tv_nsec/1000000));
 #endif
#endif

return msg;
}




static long ReadInt(char *LineX)
{
char ch;
long n = 0;

  if(LineX==NULL) return 0;

  ch=*LineX++;
  while(isspace(ch) && ch!=0)
     {
     ch=*LineX++;
     }
     
  if(!isdigit(ch)) return -1;

  while(isdigit(ch) && ch!=0)
	{
	n=10*n+(ch-'0');
	ch=*LineX++;
	}

return n;
}


#ifndef _WIN32
unsigned GetTickCount(void)
{
struct timespec tpCurr;
  clock_gettime(CLOCK_REALTIME, &tpCurr);
  return(tpCurr.tv_sec*1000+tpCurr.tv_nsec/1000000);
}
#endif



static float MinRadiation = -1;
static float MaxRadiation = -1;

/** Update RRD database. */
void UpdateRadiation(float Radiation)
{
int child_pid;
int status;
char Radiation_S[20];
static char WriteCount = 0;
unsigned tpCurr;
static unsigned tpOld = 0;

  if(MinRadiation > MaxRadiation)
    MinRadiation = MaxRadiation = Radiation;


  if(rrdupdate_PATH==NULL || RRD_DATABASE==NULL) return;


	/* Update Min and Max to allow minimisation of RRD calls. */
  WriteCount++;
  if(Radiation>=0)
  {  
    if(Radiation>MaxRadiation) MaxRadiation=Radiation;
    if(Radiation<MinRadiation || MinRadiation<0) MinRadiation=Radiation;
  }


  tpCurr = GetTickCount();
  if(labs((tpCurr - tpOld)) < MAX_RRD_RATE_MS) return;
  tpOld = tpCurr;


  if(Radiation>0) // Send Min and Max in odd and even round.
  {
    if(WriteCount & 1)
      {sprintf(Radiation_S,"N:%2.2f",MinRadiation); MinRadiation = MaxRadiation;}
    else
      {sprintf(Radiation_S,"N:%2.2f",MaxRadiation); MaxRadiation = MinRadiation;}
  }
  else	// Radiation lost, but Min and Max should be flushed.
  {
    if(MinRadiation>0)
      {sprintf(Radiation_S,"N:%2.2f",MinRadiation); MinRadiation = -1;}
    else if(MaxRadiation>0)
      {sprintf(Radiation_S,"N:%2.2f",MaxRadiation); MaxRadiation = -1;}
    else
      strcpy(Radiation_S,"N:U");		// Unknown Radiation
  }

  //printf("\n%s %s %s", rrdupdate_PATH, RRD_DATABASE, Radiation_S);

 #ifndef _WIN32
  if((child_pid = fork()) < 0)
    {
    if(LogFile)
      fprintf(LogFile,"Fork for RRUPDATE failure");
    rrdupdate_PATH=NULL;
    return;
    }

  if(child_pid == 0)	//a new-born process
  {
     // /usr/sbin/rrdupdate /var/rrd/rodgarssi.rrd N:$RODGASIGNAL
    execl(rrdupdate_PATH, rrdupdate_PATH, 
		RRD_DATABASE, Radiation_S, (char *)0);
    perror("Exec for RRUPDATE failed!");
    _exit(1);
    }

  if(Verbosity>=0) printf(" RRD-");
  waitpid(child_pid,&status,0);
  if(Verbosity>=0)
    printf((status==0)?"OK ":"FAIL ");
#else

  
#endif
}  




/** This function prints help to the text console. */
void ShowHelp(void)
{
  printf("\nUsage: smg2 [switches]"
	 "\nSwitches:"
	 "\n  -?  --help   Display this help page.");
  printf("\n  --COM_PORT dev     [%s] set SMG COM port", CHK(MODEMDEVICE));
  
  printf("\n  --rrdupdate_PATH rr_path [%s] set path to 'rrdupdate' command", CHK(rrdupdate_PATH));

  printf("\n  --RRD_DATABASE db_path [%s] set path to rrd Radiation database", CHK(RRD_DATABASE));

  printf("\n  --SMG2_STAT_FILE file [%s] SMG2 status file", CHK(SMG2_STAT_FILE));
  printf("\n  --SERIAL_TIMEOUT val [%g] Timeout for serial line in [s].", SerialTimeout_us/1000000.0);
  printf("\n  --LOG_FILE file      [%s] file with logs", CHK(LOG_FILE));
  printf("\n  --MAX_RRD_RATE_MS val [%u] Maximal allowed rate for RRD database update.", MAX_RRD_RATE_MS);
  printf("\n  -s | -v             silent or verbose mode");

  printf("\n");
}


/** This function prints text console heading. */
void Heading(void)
{
static int printed = 0;

 if(printed) return;	//print heading only once
 printed = 1;

 printf(">>>SMG2 radiation reader<<< Copyright: (C) 2013-2014 Jaroslav Fojtik"
      "\ninspired by read_rss\n");
 fflush(stdout);
}


/** Checks whether there is at least one additional argument. */
int CheckArg(int *p, int argc)
{
  if(p==NULL) return 0;
  if(++*p>=argc)
  {
    if(LogFile)
      fprintf(LogFile,"\nMissing argument!");
    return 0;  
  }

return 1;  
}


/** This procedure handles parameters and the filenames.
 * @param[in] DbHandler		Databaze connetion parameters.
 * @return	 		0 - parameters read; 1 - exit program */
int ProcessCommandLine(int argc, char *argv[])
{
  char *StrName = NULL;
  const char *name;
  int p;
  char firstch;

  for(p=1; p<argc; p++)
    {
    if(StrName) {free(StrName); StrName=NULL;}
    StrName = strdup(argv[p]);
    if(StrName=="") continue;

    firstch = StrName[0];
    if (StrName[0] == '/') StrName[0] = '-';
    name = StrName;

    if (name[0] == '-')
      {

      if(!strcmp(name, "--COM_PORT"))
        {
        if(!CheckArg(&p,argc)) continue;
	    MODEMDEVICE = argv[p];
	    continue;
        }

      if(!strcmp(name, "--SMG2_STAT_FILE"))
        {
        if(!CheckArg(&p,argc)) continue;
	    SMG2_STAT_FILE = argv[p];
	    continue;
        }

      if(!strcmp(name, "--CRTSCTS"))
        {
	    CRtsCts = CRTSCTS;
	    continue;
        }

      if(!strcmp(name, "--NO_CRTSCTS"))
        {
	    CRtsCts = 0;
	    continue;
        }

      if(!strcmp(name, "--help") || !strcmp(name, "-?") || !strcmp(name, "-help"))
	   {
	   Heading();
	   ShowHelp();
	   return(1);
	   }

      if(!strcmp(name, "--LOG_FILE"))
        {
        if(!CheckArg(&p,argc)) continue;
	    LOG_FILE = argv[p];
	    continue;
        }

      if(!strcmp(name, "--MAX_RRD_RATE_MS"))
        {
        if(!CheckArg(&p,argc)) continue;	 
	    MAX_RRD_RATE_MS = atoi(argv[p]);
	    continue;
        }

      if(!strcmp(name, "--MAX_RRD_RATE"))
        {
        if(!CheckArg(&p,argc)) continue;	 
	    MAX_RRD_RATE_MS = 1000*atof(argv[p]);
	    continue;
        }

      if(!strcmp(name, "--NO_rrdupdate") || !strcmp(name, "--NO_RRUPDATE"))
        {
	    rrdupdate_PATH = NULL;
	    continue;
        }

      if(!strcmp(name, "--SERIAL_TIMEOUT"))
        {
        if(!CheckArg(&p,argc)) continue;	 
	    SerialTimeout_us = 1000000*atof(argv[p]);
	    if(SerialTimeout_us<MINIMAL_TIMEOUT_us)
	      {
	      if(LogFile)
	      fprintf(LogFile,"\nBad serial timeout \"%g\".", SerialTimeout_us/1000000.0);
	      SerialTimeout_us=MINIMAL_TIMEOUT_us;
	      }
	   continue;
        }

      if(!strcmp(name, "--rrdupdate_PATH") || !strcmp(name, "--RRDUPDATE_PATH"))
         {
         if(!CheckArg(&p,argc)) continue;
	 rrdupdate_PATH = argv[p];
	 continue;
         }

      if(!strcmp(name, "--RRD_DATABASE"))
         {
         if(!CheckArg(&p,argc)) continue;
	 RRD_DATABASE = argv[p];
	 continue;
         }

       if(!strcmp(name, "-s"))
         {
         Verbosity--;         
	 continue;
         }

       if(!strcmp(name, "-v"))
         {
         Verbosity++;
	 continue;
         }

      StrName[0] = firstch;
      if(firstch=='-') goto UnknownArg;
      }

UnknownArg:
    Heading();
    if(LogFile)
      fprintf(LogFile,"\nUnknown argument \"%s\"!",StrName);
    SigTerm=1;
  }

  if(StrName) {free(StrName); StrName=NULL;}
return(0);
}


/*****************************************************/
/** This is a main signal handler. */
void sig_int(int signo)
{
// printf("Signal received %d!"); fflush(stdout);

 switch(signo)
    {
#ifdef SIGUSR1
      case SIGUSR1: SigUsrCount++;       // SIGUSR1 breaks a delay.
		    if(LogFile==NULL)
		      {
		      if(LOG_FILE!=NULL)
			{
			LogFile = fopen(LOG_FILE,"ab");
		        }
		      }
		    else
		      {
		      if(LogFile!=stderr)
		        {                        
		        fprintf(LogFile,"\n%s LogFile detached because of SIGUSR1 received.", FormatTime());
		        fclose(LogFile);
		        LogFile = NULL;
		        }
		      }
		    break;
#endif

#ifdef SIGUSR2
      case SIGUSR2: 
		    if(F>=0)
		      {
		      if(LogFile)
		        fprintf(LogFile,"\n%s Serial line detached because of SIGUSR2 received.", FormatTime());
		      // restore_serial_line(F, &old_tio);
		      close(F);
		      F =- 1;
		      }
		    else
		      {
		      if(MODEMDEVICE!=NULL && *MODEMDEVICE!=0)
		        {
			F = open_serial(MODEMDEVICE);
			if(F>=0)
			  {
			  if(LogFile)
	                    fprintf(LogFile,"\n%s Serial line reopened because of SIGUSR2 received.", FormatTime());
			  set_serial_line(F,BAUDRATE);
			  write(F, "Z\r", 2);		// Check for base.
			  }
			}
		      }
		    break;
#endif

      case SIGINT:
      case SIGTERM: if(SigTerm++>=2)
		      {
		      if(LogFile)
		        fprintf(LogFile,"\n%s 2nd SIGTERM has been received.",FormatTime());
		      abort();
		      }
		    else
		      {
		      if(LogFile && LogFile!=stderr)
                        fprintf(LogFile,"\n%s SIGTERM has been received - scheduling finish.",FormatTime());
                      RequestCount = 0;
		      fprintf(stderr,"\n%s SIGTERM has been received - scheduling finish.",FormatTime());
		      }
		    break;		    
      case SIGSEGV: {
		    if(LogFile)
		      fprintf(stderr,"\n%s Access violation, please report a bug.",FormatTime());
                    break;
		    }
    }
}


/** Registers all signals that are aepected to be catched in application. */
static void HookSignals(void)
{
#ifdef SIGUSR1
  signal(SIGUSR1,sig_int);
#endif
#ifdef SIGUSR2
  signal(SIGUSR2,sig_int);
#endif
  signal(SIGINT,sig_int);
  signal(SIGTERM,sig_int);
  signal(SIGSEGV,sig_int);
}


/** Read one line SMG2 report from the serial line. */
int ReadSMG2Line(unsigned *Block, T_COM F)
{
char c;
int seq;
int Timeout = SerialTimeout_us;
int bytes_read;

  seq = 0;

  while(Timeout>0 && SigTerm<=0 && seq<32)
  {
    c = 0;
    
    if((bytes_read=read_serial(F,(char *)&Block[seq],4)) <=0) break;    
    if(bytes_read<4) break;
    seq++;
  }

 if(seq<=0) return -1;
 return Block[1] & 0xFFFF;
}


const unsigned char SMG2_REQUEST[5] = {0x81, 0x15, 0xFF, 0xFF, 0x94};


/** Main worker. */
int main(int argc, char *argv[])
{
FILE *F2;
int i, OldI;
unsigned Block[32];
float Radiation;

  LogFile = stderr;
  if(ProcessCommandLine(argc,argv)!=0)
    return 1;

  Heading();

  if(LOG_FILE!=NULL)
    {
      LogFile = fopen(LOG_FILE,"ab");
      if(LogFile==NULL)
        LogFile = stderr;
      else
        fprintf(LogFile,"\n%s SMG2 governor started.",FormatTime());
    }

  if(Verbosity>0 && MODEMDEVICE!=NULL)
    printf("\n Opening serial line \"%s\".",CHK(MODEMDEVICE));
  F = open_serial(MODEMDEVICE);
  if(F<0)
    {
    if(LogFile)
      fprintf(LogFile, "\n%s Cannot open serial line (%s).", FormatTime(), MODEMDEVICE);
    return -1;
    }

  HookSignals();		/* Hook signals like CTRL+C. */

  set_serial_line(F,BAUDRATE);
  i = 0;
  OldI = ~0;
  RequestCount = 0;

  write_serial(F, SMG2_REQUEST,5);
  while(SigTerm<=0)
    {
    i = ReadSMG2Line(Block,F);
    if(i<0)
      {
      if(Verbosity>1) printf("\n No response from device, waiting.");
      do
        {        
        Sleep(500); 
        if(Verbosity>0)
          {
          if(RequestCount<=0) 
            printf(" (REQ sent)");
          else if(RequestCount==1) 
            printf("    1");
          else
            printf("\b\b\b\b\b%5u",RequestCount);
          fflush(stdout);
          }
        RequestCount++;
        } while(write_serial(F,SMG2_REQUEST,5)<=0 && SigTerm<=0);
      continue;
      }

    if(i==OldI) continue;

    if(Block[0]!= 0x820) continue;

    OldI = i;			/* Measurement index. */
    RequestCount = 0;


    Radiation = Block[2] / 100.0f;    

    // Block[0] Packet identification
    // Block[1] Measurement sequence number
    // Block[2] Current radiation
    // Block[3] Average Bq/hour
    // Block[4] Average Bq/day

    printf("\r\n Radiation is %2.2f < %2.2f < %2.2f uSv    ", MinRadiation, Radiation, MaxRadiation);
    if(Verbosity>=1)
    {
      printf("\r\n");
      for(i=0; i<32; i++)
      {
         if(i==0 || i==1 || i==30 || i==31) 
           printf("%X ",Block[i]);
         else
           printf("%u ",Block[i]);
      }
    }
    fflush(stdout);
    
    UpdateRadiation(Radiation);

    if(SMG2_STAT_FILE)
      {
      F2 = fopen(SMG2_STAT_FILE,"wb");
      if(F2)
        {
        fprintf(F2,"Radiation is %2.2f uSv\n", Radiation);
        fclose(F2);
        }
      }
    
    }

	/* Exit program. */    
  if(F>=0)
    {
    restore_serial_line(F);
    close_serial_line(F);
    F = INVALID_HANDLE;
    }


  if(SMG2_STAT_FILE)
    {
    F2 = fopen(SMG2_STAT_FILE,"ab");
    if(F2)
      {
      fprintf(F2,"\nSMG2 governor is not active.");
      fprintf(F2,"\nLast response:\n%2.2f",Radiation);
      fprintf(F2,"\n");
      fclose(F2);
      }
    }
 
  if(LogFile!=stderr && LogFile!=NULL)
  {
     fclose(LogFile);
     LogFile = NULL;
  }

return 0;
}