/* SMG2 serial linux reader; (C)2012-2014 Jaroslav Fojtik, Licence LGPL. */


#include <stdio.h>          /* Standard input/output definitions */
#include <stdlib.h>
#include <string.h>

#include <fcntl.h>          /* File control definitions */


#ifndef _WIN32
  #include <termios.h>
  #define closesocket(_s) close(_s)  
#endif

#include "serial.h"



static struct termios oldtio;


//#define _POSIX_SOURCE 1 /* POSIX compliant source */

/* baudrate settings are defined in <asm/termbits.h>, which is
   included by <termios.h> */
extern int CRtsCts;



T_COM open_serial(const char *devname)
{
  T_COM fd;
	/* Open modem device for reading and writing and not as controlling tty
	   because we don't want to get killed if linenoise sends CTRL-C. */
  fd = open(devname, O_RDWR | O_NOCTTY | O_NDELAY); 
  if(fd < 0) { perror(devname); }
  return fd;
}


int set_serial_line(T_COM fd, int BAUDRATE)
{
struct termios newtio;

    switch(BAUDRATE)
    {
      case 0:      BAUDRATE = B0; break;
      case 50:     BAUDRATE = B50; break;
      case 75:     BAUDRATE = B75; break;
      case 110:    BAUDRATE = B110; break;
      case 134:    BAUDRATE = B134; break;
      case 150:    BAUDRATE = B150; break;
      case 200:    BAUDRATE = B200; break;
      case 300:    BAUDRATE = B300; break;
      case 600:    BAUDRATE = B600; break;
      case 1200:   BAUDRATE = B1200; break;
      case 1800:   BAUDRATE = B1800; break;
      case 2400:   BAUDRATE = B2400; break;
      case 4800:   BAUDRATE = B4800; break;
      case 9600:   BAUDRATE = B9600; break;
      case 19200:  BAUDRATE = B19200; break;
      case 38400:  BAUDRATE = B38400; break;
      case 57600:  BAUDRATE = B57600; break;
      case 115200: BAUDRATE = B115200; break;
      case 230400: BAUDRATE = B230400; break;
      case 460800: BAUDRATE = B460800; break;
      case 500000: BAUDRATE = B500000; break;
      case 576000: BAUDRATE = B576000; break;
    }

//  if(oldtio)
    tcgetattr(fd, &oldtio); /* save current serial port settings */

  memset(&newtio, 0, sizeof(newtio)); /* clear struct for new port settings */

    /* BAUDRATE: Set bps rate. You could also use cfsetispeed and cfsetospeed.
       CRTSCTS : output hardware flow control (only used if the cable has
	            all necessary lines. See sect. 7 of Serial-HOWTO)
       CS8     : 8n1 (8bit,no parity,1 stopbit)
       CLOCAL  : local connection, no modem contol
       CREAD   : enable receiving characters
       HUPCL   : Lower modem control lines after last process closes the device (hang up).  */
  newtio.c_cflag = BAUDRATE | CRtsCts | CS8 | CLOCAL | CREAD; //  | HUPCL;
	 
    /* IGNPAR  : ignore bytes with parity errors
       ICRNL   : map CR to NL (otherwise a CR input on the other computer
                 will not terminate input)
                 otherwise make device raw (no other input processing) */
  newtio.c_iflag = IGNPAR | ICRNL;
         
    /* Raw output. */
  newtio.c_oflag = 0;
         
    /* ICANON  : enable canonical input
	  disable all echo functionality, and don't send signals to calling program */
  newtio.c_lflag = 0; 		// !!!!!! ICANON;
         
    /* initialize all control characters 
    default values can be found in /usr/include/termios.h, and are given
    in the comments, but we don't need them here */
  newtio.c_cc[VINTR]    = 0;     /* Ctrl-c */ 
  newtio.c_cc[VQUIT]    = 0;     /* Ctrl-\ */
  newtio.c_cc[VERASE]   = 0;     /* del */
  newtio.c_cc[VKILL]    = 0;     /* @ */
  newtio.c_cc[VEOF]     = 4;     /* Ctrl-d */
  newtio.c_cc[VTIME]    = 0;     /* inter-character timer unused */
  newtio.c_cc[VMIN]     = 1;     /* blocking read until 1 character arrives */
  newtio.c_cc[VSWTC]    = 0;     /* '\0' */
  newtio.c_cc[VSTART]   = 0;     /* Ctrl-q */ 
  newtio.c_cc[VSTOP]    = 0;     /* Ctrl-s */
  newtio.c_cc[VSUSP]    = 0;     /* Ctrl-z */
  newtio.c_cc[VEOL]     = 0;     /* '\0' */
  newtio.c_cc[VREPRINT] = 0;     /* Ctrl-r */
  newtio.c_cc[VDISCARD] = 0;     /* Ctrl-u */
  newtio.c_cc[VWERASE]  = 0;     /* Ctrl-w */
  newtio.c_cc[VLNEXT]   = 0;     /* Ctrl-v */
  newtio.c_cc[VEOL2]    = 0;     /* '\0' */

    /* now clean the modem line and activate the settings for the port. */
  tcflush(fd, TCIFLUSH);
  tcsetattr(fd,TCSANOW, &newtio);

  return 0;
}


int restore_serial_line(T_COM fd)
{
	/* restore the old port settings */
  tcsetattr(fd, TCSANOW, &oldtio);
  return 0;
}


void close_serial_line(T_COM HCom)
{
  close(HCom);
}


int write_serial(T_COM HCom, const char *DATA, unsigned size)
{
  return write(HCom, DATA, size);
}


int read_serial(T_COM HCom, char *DATA, unsigned size)
{
  return read(HCom, DATA, size);
}



